
from .internal import (
    _is_python_2_7_or_later,
    _is_python_3_0_or_later,
)


__all__ = [

    'UNSPECIFIED',
    'VIOLATION',
    'ALERT',
    'CRITICAL',
    'FAILURE',
    'WARNING',
    'NOTICE',
    'INFORMATIONAL',
    'DEBUG0',
    'DEBUG1',
    'DEBUG2',
    'DEBUG3',
    'DEBUG4',
    'DEBUG5',
    'TRACE',
    'BENCHMARK',

    'FAIL',
    'WARN',
    'INFO',

    'parse_verbosity',
    'severity_to_string',
]

UNSPECIFIED     =   0
VIOLATION       =   1
ALERT           =   2
CRITICAL        =   3
FAILURE         =   4
WARNING         =   5
NOTICE          =   6
INFORMATIONAL   =   7
DEBUG0          =   8
DEBUG1          =   9
DEBUG2          =   10
DEBUG3          =   11
DEBUG4          =   12
DEBUG5          =   13
TRACE           =   14
BENCHMARK       =   15

DEBUG           =   DEBUG5
FAIL            =   FAILURE
WARN            =   WARNING
INFO            =   INFORMATIONAL

_STOCK_SEVERITY_LEVELS = {

    VIOLATION       :   'VIOLATION',
    ALERT           :   'ALERT',
    CRITICAL        :   'CRITICAL',
    FAILURE         :   'FAILURE',
    WARNING         :   'WARNING',
    NOTICE          :   'NOTICE',
    INFORMATIONAL   :   'INFORMATIONAL',
    DEBUG0          :   'DEBUG0',
    DEBUG1          :   'DEBUG1',
    DEBUG2          :   'DEBUG2',
    DEBUG3          :   'DEBUG3',
    DEBUG4          :   'DEBUG4',
    DEBUG5          :   'DEBUG5',
    TRACE           :   'TRACE',
    BENCHMARK       :   'BENCHMARK',
}

_STRINGS_RECOGNISABLE_AS_SEVERITY_LEVELS = {

    # actual values
    'VIOLATION'                 :   VIOLATION,
    'ALERT'                     :   ALERT,
    'CRITICAL'                  :   CRITICAL,
    'FAILURE'                   :   FAILURE,
    'WARNING'                   :   WARNING,
    'NOTICE'                    :   NOTICE,
    'INFORMATIONAL'             :   INFORMATIONAL,
    'DEBUG0'                    :   DEBUG0,
    'DEBUG1'                    :   DEBUG1,
    'DEBUG2'                    :   DEBUG2,
    'DEBUG3'                    :   DEBUG3,
    'DEBUG4'                    :   DEBUG4,
    'DEBUG5'                    :   DEBUG5,
    'TRACE'                     :   TRACE,
    'BENCHMARK'                 :   BENCHMARK,

    # stock shorthands
    'FAIL'                      :   FAILURE,
    'WARN'                      :   WARNING,
    'INFO'                      :   INFORMATIONAL,

    # de-facto public synonyms
    'EMERGENCY'                 :   VIOLATION,
    'ERROR'                     :   FAILURE,
    'DEBUG'                     :   DEBUG5,

    # de-facto public shorthands
    'ALRT'                      :   ALERT,
    'CRIT'                      :   CRITICAL,
    'ERR'                       :   FAILURE,
    'TRC'                       :   TRACE,
    'BENCH'                     :   BENCHMARK,
}

if _is_python_3_0_or_later():

    _INTEGER_TYPES = (int, )
else:

    _INTEGER_TYPES = (int, long, )


def _parse_verbosity(
    v,
    strict_case_comparison,
):
    """
    Parses the verbosity from `v`.

    Parameters
    ----------
    v : str | int | *
        The variable from which to parse the verbosity. If an integer, is taken as is; if not a string then converted to a string, and in either case is then subject to parsing
    strict_case_comparison : bool
        Specifies whether any string comparison should be strict or permissive

    Returns
    -------
    int
        The severity level

    Raises
    ------
    ValueError
        If the string form of `v` does not contain a recognisable severity level
    """

    if isinstance(v, _INTEGER_TYPES):

        return v

    if isinstance(v, str):

        s = v.strip()
    else:

        s = str(v)

    severity = _STRINGS_RECOGNISABLE_AS_SEVERITY_LEVELS.get(s, None)
    if severity == None and not strict_case_comparison:

        s_upper = s.upper()

        severity = _STRINGS_RECOGNISABLE_AS_SEVERITY_LEVELS.get(s_upper, None)

    if severity != None:

        return severity

    try:

        return int(s)
    except ValueError as x:

        raise ValueError("could not recognise value '%s' as a severity" % (v))


if _is_python_2_7_or_later():

    def parse_verbosity(
        v,
        **kwargs,
    ):
        """
        Attempts to parse the verbosity from `v`.

        Parameters
        ----------
        v : str | int | *
            The variable from which to parse the verbosity. If an integer, is taken as is; if not a string is converted to a string, and in either case is then subject to parsing
        kwargs : dict
            Keyword arguments. Currently only `strict_case_comparison` is recognised, which defaults to `False` if not specified by the caller

        Returns
        -------
        int
            The severity level

        Raises
        ------
        ValueError
            If the string form of `v` does not contain a recognisable severity level
        """

        strict_case_comparison = kwargs.get('strict_case_comparison', False)


        return _parse_verbosity(
            v,
            strict_case_comparison=strict_case_comparison,
        )
else:

    def parse_verbosity(
        v,
    ):
        """
        Attempts to parse the verbosity from `v`.

        Parameters
        ----------
        v : str | int | *
            The variable from which to parse the verbosity. If an integer, is taken as is; if not a string is converted to a string, and in either case is then subject to parsing

        Returns
        -------
        int
            The severity level

        Raises
        ------
        ValueError
            If the string form of `v` does not contain a recognisable severity level
        """

        return _parse_verbosity(
            v,
            False, # strict_case_comparison
        )


def severity_to_string(severity):
    """
    Converts a severity to its string form

    Parameters
    ----------
    severity : int
        The severity associated with the message

    Returns
    -------
    str
        The string corresponding to the severity
    """

    s = _STOCK_SEVERITY_LEVELS.get(severity, None)

    if s is None:

        if isinstance(severity, _INTEGER_TYPES):

            s = str(severity)
        else:

            s = "%s (%s)" % (str(severity), type(severity))

        return "<Severity: %s>" % s
    else:

        return s


# ############################## end of file ############################# #




#! /usr/bin/env python3

import diagnosticism as d

import sys


def func1():

    #sys.stderr.write("locals(%s)=%s\n" % (type

    #d.trace()
    pass


def func2(x, y):

    d.trace()


def func3(x, y, **options):

    d.trace()

    z = x * y

    return z


def func4(x, y, *args):

    d.trace()


def func5(x, y, *args, **options):

    d.trace()


class Thing:


    def __init__(self):

        d.trace()

        pass


    def some_method(self, x, y, z):

        d.trace()

        pass


d.enable_tracing(('DIAGNOSTICISM_ENABLE_TRACING', 'ENABLE_TRACING'), True)


d.trace()

func1()

func2('abc', 12)

func3('abc', 12)
func3('abc', 12, pos='left')

func4('abc', 12)
func4('abc', 12, 13, 14, 15)

func5('abc', 12)
func5('abc', 12, 13, 14, 15)
func5('abc', 12, pos='left', length=10)
func5('abc', 12, 13, 14, 15, pos='left', length=10)

thing = Thing()
thing.some_method(-1, 0, +1)




#! /usr/bin/env python3

from diagnosticism.contingent_reporting import (
    abort,
    conrep,
    report,
    set_default_trailing_prompt,
)
from diagnosticism.program_name import set_program_name

import unittest
from unittest.mock import patch

try:

    from StringIO import StringIO
except ImportError:

    from io import StringIO


class ConRep_tester(unittest.TestCase):


    def test_with_program_name(self):

        with patch('sys.stderr', new=StringIO()) as fake_stderr:

            set_program_name('myprog1')

            conrep('string-1', show_program_name=True)

            self.assertEqual("myprog1: string-1\n", fake_stderr.getvalue())


    def test_without_program_name(self):

        with patch('sys.stderr', new=StringIO()) as fake_stderr:

            conrep('string-1', show_program_name=False)

            self.assertEqual("string-1\n", fake_stderr.getvalue())


class Report_tester(unittest.TestCase):


    def test_with_program_name(self):

        with patch('sys.stderr', new=StringIO()) as fake_stderr:

            set_program_name('myprog1')

            report('string-1', show_program_name=True)

            self.assertEqual("myprog1: string-1\n", fake_stderr.getvalue())


    def test_without_program_name(self):

        with patch('sys.stderr', new=StringIO()) as fake_stderr:

            report('string-1', show_program_name=False)

            self.assertEqual("string-1\n", fake_stderr.getvalue())


    def test__report__WITH_file_PARAM_AND_show_program_name_False(self):

        file = StringIO()

        set_program_name('myprog3')

        report('string-3', file=file, show_program_name=False)

        result = file.getvalue()

        self.assertEqual("string-3\n", result)


    def test__report__WITH_file_PARAM_AND_show_program_name_True(self):

        file = StringIO()

        set_program_name('myprog4')

        report('string-4', file=file, show_program_name=True)

        result = file.getvalue()

        self.assertEqual("myprog4: string-4\n", result)


class Abort_tester(unittest.TestCase):


    def test_default(self):

        with patch('sys.stderr', new=StringIO()) as fake_stderr:

            set_program_name('myprog1')

            abort('over and out!', do_exit=False)

            self.assertEqual("myprog1: over and out!\n", fake_stderr.getvalue())


    def test_explicit_trailing_prompt(self):

        with patch('sys.stderr', new=StringIO()) as fake_stderr:

            set_program_name('myprog1')

            abort('over and out!', do_exit=False, trailing_prompt='get over yourself!')

            self.assertEqual("myprog1: over and out!; get over yourself!\n", fake_stderr.getvalue())


    def test_stock_trailing_prompt(self):

        with patch('sys.stderr', new=StringIO()) as fake_stderr:

            set_program_name('myprog1')

            abort('over and out!', do_exit=False, trailing_prompt=True)

            self.assertEqual("myprog1: over and out!; use --help for usage\n", fake_stderr.getvalue())


    def test_set_default_trailing_prompt_1(self):

        with patch('sys.stderr', new=StringIO()) as fake_stderr:

            set_program_name('myprog1')

            set_default_trailing_prompt('ok, now')

            abort('over and out!', do_exit=False, trailing_prompt=True)

            set_default_trailing_prompt(None)

            self.assertEqual("myprog1: over and out!; ok, now\n", fake_stderr.getvalue())


    def test_set_default_trailing_prompt_2(self):

        with patch('sys.stderr', new=StringIO()) as fake_stderr:

            set_program_name('myprog1')

            set_default_trailing_prompt('ok, now')

            abort('over and out!', do_exit=False, trailing_prompt=False)

            set_default_trailing_prompt(None)

            self.assertEqual("myprog1: over and out!\n", fake_stderr.getvalue())


if '__main__' == __name__:

    unittest.main()




#! /usr/bin/env python3

from diagnosticism import (
    DOOMGram,
    DOOMScope,
)

import random
import time


def main():

    dg = DOOMGram()

    for _ in range(1000):

        r = random.uniform(1, 1000)

        d = r / 1_000_000

        with DOOMScope(dg):

            time.sleep(d)

    print("dg:", dg)


if __name__ == "__main__":

    main()



#! /usr/bin/env python3

from diagnosticism import (
    file,
    func,
    line,
    fileline,
    filelinefunc,
)

def f():

    print("  `file()`:", file())

    print("  `file(basename=True)`:", file(basename=True))

    print("  `line()`:", line())

    print("  `func()`:", func())

    print("  `fileline()`:", fileline())

    print("  `fileline(basename=True)`:", fileline(basename=True))

    print("  `filelinefunc()`:", filelinefunc())

    print("  `filelinefunc(basename=True)`:", filelinefunc(basename=True))


if __name__ == "__main__":

    print()
    print("in `f()`:")

    f()

    print()
    print("in `main()`:")

    print("  `file()`:", file())

    print("  `file(basename=True)`:", file(basename=True))

    print("  `line()`:", line())

    print("  `func()`:", func())

    print("  `fileline()`:", fileline())

    print("  `fileline(basename=True)`:", fileline(basename=True))

    print("  `filelinefunc()`:", filelinefunc())

    print("  `filelinefunc(basename=True)`:", filelinefunc(basename=True))




#! /usr/bin/env python3

from diagnosticism import DOOMGram

import unittest


class DOOMGram_tester(unittest.TestCase):

    def test_empty(self):

        dg = DOOMGram()

        self.assertEqual(0, dg.event_count())

        self.assertEqual(0, dg.total_event_time_ns())
        self.assertIsNone(dg.min_event_time_ns())
        self.assertIsNone(dg.max_event_time_ns())

        self.assertEqual(0, dg.num_events_in_1ns())
        self.assertEqual(0, dg.num_events_in_10ns())
        self.assertEqual(0, dg.num_events_in_100ns())
        self.assertEqual(0, dg.num_events_in_1us())
        self.assertEqual(0, dg.num_events_in_10us())
        self.assertEqual(0, dg.num_events_in_100us())
        self.assertEqual(0, dg.num_events_in_1ms())
        self.assertEqual(0, dg.num_events_in_10ms())
        self.assertEqual(0, dg.num_events_in_100ms())
        self.assertEqual(0, dg.num_events_in_1s())
        self.assertEqual(0, dg.num_events_in_10s())
        self.assertEqual(0, dg.num_events_ge_100s())

        self.assertEqual("____________", dg.to_strip())
        self.assertEqual("____________", str(dg))
        self.assertEqual("DOOMGram(event_count=0, total_event_time_ns=0, min_event_time_ns=None, max_event_time_ns=None, num_events_in_1ns=0, num_events_in_10ns=0, num_events_in_100ns=0, num_events_in_1us=0, num_events_in_10us=0, num_events_in_100us=0, num_events_in_1ms=0, num_events_in_10ms=0, num_events_in_100ms=0, num_events_in_1s=0, num_events_in_10s=0, num_events_ge_100s=0)", repr(dg))

    def test_single_timing_event(self):

        dg = DOOMGram()

        dg.push_event_time_ms(13)

        self.assertEqual(1, dg.event_count())

        self.assertEqual(13000000, dg.total_event_time_ns())
        self.assertEqual(13000000, dg.min_event_time_ns())
        self.assertEqual(13000000, dg.max_event_time_ns())

        self.assertEqual(0, dg.num_events_in_1ns())
        self.assertEqual(0, dg.num_events_in_10ns())
        self.assertEqual(0, dg.num_events_in_100ns())
        self.assertEqual(0, dg.num_events_in_1us())
        self.assertEqual(0, dg.num_events_in_10us())
        self.assertEqual(0, dg.num_events_in_100us())
        self.assertEqual(0, dg.num_events_in_1ms())
        self.assertEqual(1, dg.num_events_in_10ms())
        self.assertEqual(0, dg.num_events_in_100ms())
        self.assertEqual(0, dg.num_events_in_1s())
        self.assertEqual(0, dg.num_events_in_10s())
        self.assertEqual(0, dg.num_events_ge_100s())

        self.assertEqual("_______a____", dg.to_strip())
        self.assertEqual("_______a____", str(dg))

    def test_zero_time_events(self):

        dg = DOOMGram()

        dg.push_event_time_ns(0)
        dg.push_event_time_us(0)
        dg.push_event_time_ms(0)
        dg.push_event_time_s(0)

        self.assertEqual(4, dg.event_count())

        self.assertEqual(0, dg.total_event_time_ns())
        self.assertEqual(0, dg.min_event_time_ns())
        self.assertEqual(0, dg.max_event_time_ns())

        self.assertEqual(0, dg.num_events_in_1ns())
        self.assertEqual(0, dg.num_events_in_10ns())
        self.assertEqual(0, dg.num_events_in_100ns())
        self.assertEqual(0, dg.num_events_in_1us())
        self.assertEqual(0, dg.num_events_in_10us())
        self.assertEqual(0, dg.num_events_in_100us())
        self.assertEqual(0, dg.num_events_in_1ms())
        self.assertEqual(0, dg.num_events_in_10ms())
        self.assertEqual(0, dg.num_events_in_100ms())
        self.assertEqual(0, dg.num_events_in_1s())
        self.assertEqual(0, dg.num_events_in_10s())
        self.assertEqual(0, dg.num_events_ge_100s())

        self.assertEqual("____________", dg.to_strip())
        self.assertEqual("____________", str(dg))

    def test_uniform_spread_timings_1(self):

        dg = DOOMGram()

        dg.push_event_time_ns(    9)
        dg.push_event_time_ns(   80)
        dg.push_event_time_ns(  700)
        dg.push_event_time_us(    6)
        dg.push_event_time_us(   50)
        dg.push_event_time_us(  400)
        dg.push_event_time_ms(    3)
        dg.push_event_time_ms(   20)
        dg.push_event_time_ms(  100)
        dg.push_event_time_s(     9)
        dg.push_event_time_s(    80)
        dg.push_event_time_s(   700)


        self.assertEqual(12, dg.event_count())

        self.assertEqual(789123456789, dg.total_event_time_ns())
        self.assertEqual(9, dg.min_event_time_ns())
        self.assertEqual(700_000_000_000, dg.max_event_time_ns())

        self.assertEqual(1, dg.num_events_in_1ns())
        self.assertEqual(1, dg.num_events_in_10ns())
        self.assertEqual(1, dg.num_events_in_100ns())
        self.assertEqual(1, dg.num_events_in_1us())
        self.assertEqual(1, dg.num_events_in_10us())
        self.assertEqual(1, dg.num_events_in_100us())
        self.assertEqual(1, dg.num_events_in_1ms())
        self.assertEqual(1, dg.num_events_in_10ms())
        self.assertEqual(1, dg.num_events_in_100ms())
        self.assertEqual(1, dg.num_events_in_1s())
        self.assertEqual(1, dg.num_events_in_10s())
        self.assertEqual(1, dg.num_events_ge_100s())

        self.assertEqual("aaaaaaaaaaaa", dg.to_strip())
        self.assertEqual("aaaaaaaaaaaa", str(dg))

    def test_uniform_spread_timings_2(self):

        dg = DOOMGram()

        dg.push_event_time_ns(     9)
        dg.push_event_time_ns(    80)
        dg.push_event_time_ns(   700)
        dg.push_event_time_ns(  6000)
        dg.push_event_time_ns( 50000)
        dg.push_event_time_ns(400000)
        dg.push_event_time_ms(     3)
        dg.push_event_time_ms(    20)
        dg.push_event_time_ms(   100)
        dg.push_event_time_ms(  9000)
        dg.push_event_time_ms( 80000)
        dg.push_event_time_ms(700000)

        self.assertEqual(12, dg.event_count())

        self.assertEqual(789123456789, dg.total_event_time_ns())
        self.assertEqual(9, dg.min_event_time_ns())
        self.assertEqual(700_000_000_000, dg.max_event_time_ns())

        self.assertEqual(1, dg.num_events_in_1ns())
        self.assertEqual(1, dg.num_events_in_10ns())
        self.assertEqual(1, dg.num_events_in_100ns())
        self.assertEqual(1, dg.num_events_in_1us())
        self.assertEqual(1, dg.num_events_in_10us())
        self.assertEqual(1, dg.num_events_in_100us())
        self.assertEqual(1, dg.num_events_in_1ms())
        self.assertEqual(1, dg.num_events_in_10ms())
        self.assertEqual(1, dg.num_events_in_100ms())
        self.assertEqual(1, dg.num_events_in_1s())
        self.assertEqual(1, dg.num_events_in_10s())
        self.assertEqual(1, dg.num_events_ge_100s())

        self.assertEqual("aaaaaaaaaaaa", dg.to_strip())
        self.assertEqual("aaaaaaaaaaaa", str(dg))

    def test_uniform_spread_timings_3(self):

        dg = DOOMGram()

        dg.push_event_time_ns(            9)
        dg.push_event_time_ns(           80)
        dg.push_event_time_ns(          700)
        dg.push_event_time_ns(         6000)
        dg.push_event_time_ns(        50000)
        dg.push_event_time_ns(       400000)
        dg.push_event_time_ns(      3000000)
        dg.push_event_time_ns(     20000000)
        dg.push_event_time_ns(    100000000)
        dg.push_event_time_ns(   9000000000)
        dg.push_event_time_ns(  80000000000)
        dg.push_event_time_ns( 700000000000)

        self.assertEqual(12, dg.event_count())

        self.assertEqual(789123456789, dg.total_event_time_ns())
        self.assertEqual(9, dg.min_event_time_ns())
        self.assertEqual(700_000_000_000, dg.max_event_time_ns())

        self.assertEqual(1, dg.num_events_in_1ns())
        self.assertEqual(1, dg.num_events_in_10ns())
        self.assertEqual(1, dg.num_events_in_100ns())
        self.assertEqual(1, dg.num_events_in_1us())
        self.assertEqual(1, dg.num_events_in_10us())
        self.assertEqual(1, dg.num_events_in_100us())
        self.assertEqual(1, dg.num_events_in_1ms())
        self.assertEqual(1, dg.num_events_in_10ms())
        self.assertEqual(1, dg.num_events_in_100ms())
        self.assertEqual(1, dg.num_events_in_1s())
        self.assertEqual(1, dg.num_events_in_10s())
        self.assertEqual(1, dg.num_events_ge_100s())

        self.assertEqual("aaaaaaaaaaaa", dg.to_strip())
        self.assertEqual("aaaaaaaaaaaa", str(dg))

    def test_uniform_spread_timings_4(self):

        dg = DOOMGram()

        dg.push_event_time_us(         6)
        dg.push_event_time_us(        50)
        dg.push_event_time_us(       400)
        dg.push_event_time_us(      3000)
        dg.push_event_time_us(     20000)
        dg.push_event_time_us(    100000)
        dg.push_event_time_us(   9000000)
        dg.push_event_time_us(  80000000)
        dg.push_event_time_us( 700000000)

        self.assertEqual(9, dg.event_count())

        self.assertEqual(789123456000, dg.total_event_time_ns())
        self.assertEqual(6_000, dg.min_event_time_ns())
        self.assertEqual(700_000_000_000, dg.max_event_time_ns())

        self.assertEqual(0, dg.num_events_in_1ns())
        self.assertEqual(0, dg.num_events_in_10ns())
        self.assertEqual(0, dg.num_events_in_100ns())
        self.assertEqual(1, dg.num_events_in_1us())
        self.assertEqual(1, dg.num_events_in_10us())
        self.assertEqual(1, dg.num_events_in_100us())
        self.assertEqual(1, dg.num_events_in_1ms())
        self.assertEqual(1, dg.num_events_in_10ms())
        self.assertEqual(1, dg.num_events_in_100ms())
        self.assertEqual(1, dg.num_events_in_1s())
        self.assertEqual(1, dg.num_events_in_10s())
        self.assertEqual(1, dg.num_events_ge_100s())

        self.assertEqual("___aaaaaaaaa", dg.to_strip())
        self.assertEqual("___aaaaaaaaa", str(dg))

    def test_several_distinct_timings(self):

        dg = DOOMGram()

        dg.push_event_time_ns(  23)
        dg.push_event_time_ns(  10)
        dg.push_event_time_us(   7)
        dg.push_event_time_us(   7)
        dg.push_event_time_us(  89)
        dg.push_event_time_ms( 248)
        dg.push_event_time_s(    5)
        dg.push_event_time_s(  309)

        self.assertEqual(8, dg.event_count())

        self.assertEqual(314248103033, dg.total_event_time_ns())
        self.assertEqual(10, dg.min_event_time_ns())
        self.assertEqual(309_000_000_000, dg.max_event_time_ns())

        self.assertEqual(0, dg.num_events_in_1ns())
        self.assertEqual(2, dg.num_events_in_10ns())
        self.assertEqual(0, dg.num_events_in_100ns())
        self.assertEqual(2, dg.num_events_in_1us())
        self.assertEqual(1, dg.num_events_in_10us())
        self.assertEqual(0, dg.num_events_in_100us())
        self.assertEqual(0, dg.num_events_in_1ms())
        self.assertEqual(0, dg.num_events_in_10ms())
        self.assertEqual(1, dg.num_events_in_100ms())
        self.assertEqual(1, dg.num_events_in_1s())
        self.assertEqual(0, dg.num_events_in_10s())
        self.assertEqual(1, dg.num_events_ge_100s())

        self.assertEqual("_a_aa___aa_a", dg.to_strip())
        self.assertEqual("_a_aa___aa_a", str(dg))

    def test_several_intersecting_timings(self):

        dg = DOOMGram()

        dg.push_event_time_ns(   11)
        dg.push_event_time_ns(   19)
        dg.push_event_time_ns(   19)
        dg.push_event_time_us(    7)
        dg.push_event_time_us(    7)
        dg.push_event_time_us(   89)
        dg.push_event_time_ms(  248)
        dg.push_event_time_ms(4_321)
        dg.push_event_time_s(     5)
        dg.push_event_time_s(   309)

        self.assertEqual(10, dg.event_count())

        self.assertEqual(318569103049, dg.total_event_time_ns())
        self.assertEqual(11, dg.min_event_time_ns())
        self.assertEqual(309_000_000_000, dg.max_event_time_ns())

        self.assertEqual(0, dg.num_events_in_1ns())
        self.assertEqual(3, dg.num_events_in_10ns())
        self.assertEqual(0, dg.num_events_in_100ns())
        self.assertEqual(2, dg.num_events_in_1us())
        self.assertEqual(1, dg.num_events_in_10us())
        self.assertEqual(0, dg.num_events_in_100us())
        self.assertEqual(0, dg.num_events_in_1ms())
        self.assertEqual(0, dg.num_events_in_10ms())
        self.assertEqual(1, dg.num_events_in_100ms())
        self.assertEqual(2, dg.num_events_in_1s())
        self.assertEqual(0, dg.num_events_in_10s())
        self.assertEqual(1, dg.num_events_ge_100s())

        self.assertEqual("_a_aa___aa_a", dg.to_strip())
        self.assertEqual("_a_aa___aa_a", str(dg))

    def test_many_cumulative_timings(self):

        dg = DOOMGram()

        [dg.push_event_time_ns(1) for _ in range(0,10000)]

        [dg.push_event_time_us(1) for _ in range(0,1000)]

        [dg.push_event_time_ms(1) for _ in range(0,100)]

        [dg.push_event_time_s(1) for _ in range(0,10)]

        self.assertEqual(11110, dg.event_count())

        self.assertEqual(10101010000, dg.total_event_time_ns())
        self.assertEqual(1, dg.min_event_time_ns())
        self.assertEqual(1_000_000_000, dg.max_event_time_ns())

        self.assertEqual(10_000, dg.num_events_in_1ns())
        self.assertEqual(0, dg.num_events_in_10ns())
        self.assertEqual(0, dg.num_events_in_100ns())
        self.assertEqual(1_000, dg.num_events_in_1us())
        self.assertEqual(0, dg.num_events_in_10us())
        self.assertEqual(0, dg.num_events_in_100us())
        self.assertEqual(100, dg.num_events_in_1ms())
        self.assertEqual(0, dg.num_events_in_10ms())
        self.assertEqual(0, dg.num_events_in_100ms())
        self.assertEqual(10, dg.num_events_in_1s())
        self.assertEqual(0, dg.num_events_in_10s())
        self.assertEqual(0, dg.num_events_ge_100s())

        self.assertEqual("e__d__c__b__", dg.to_strip())
        self.assertEqual("e__d__c__b__", str(dg))
        self.assertEqual("*--d--c--b--", dg.to_strip(range='abcd', zero='-'))
        self.assertEqual("#  #  c  b  ", dg.to_strip(range='abc', zero=' ', overflow='#'))
        self.assertEqual("DOOMGram(event_count=11110, total_event_time_ns=10101010000, min_event_time_ns=1, max_event_time_ns=1000000000, num_events_in_1ns=10000, num_events_in_10ns=0, num_events_in_100ns=0, num_events_in_1us=1000, num_events_in_10us=0, num_events_in_100us=0, num_events_in_1ms=100, num_events_in_10ms=0, num_events_in_100ms=0, num_events_in_1s=10, num_events_in_10s=0, num_events_ge_100s=0)", repr(dg))


# ############################## end of file ############################# #




#! /usr/bin/env python3

from diagnosticism.severity import *

import unittest
from unittest.mock import patch

import re
import sys

try:

    from StringIO import StringIO
except ImportError:

    from io import StringIO


INTEGER_LEVELS = [ -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, ]

NAMED_LEVELS = {

    "Warn"          :   WARNING,
    "Warning"       :   WARNING,
    "WARN"          :   WARNING,
    "WARNING"       :   WARNING,
    "warn"          :   WARNING,
    "warning"       :   WARNING,
}


class ParseSeverity_tester(unittest.TestCase):


    def test_integer_levels_as_int(self):

        for i in INTEGER_LEVELS:

            expected = i
            actual = parse_verbosity(i)

            self.assertEqual(expected, actual)


    def test_integer_levels_as_str(self):

        for i in INTEGER_LEVELS:

            expected = i
            actual = parse_verbosity(str(i))

            self.assertEqual(expected, actual)


    def test_integer_levels_as_str_with_padding(self):

        for i in INTEGER_LEVELS:

            expected = i
            actual = parse_verbosity("\t%d " % (i))

            self.assertEqual(expected, actual)


    def test_named_levels(self):

        for (s, l) in NAMED_LEVELS.items():

            expected = l
            actual = parse_verbosity(s)

            self.assertEqual(expected, actual)


    def test_named_levels_with_padding(self):

        for (s, l) in NAMED_LEVELS.items():

            expected = l
            actual = parse_verbosity("\t%s " % (s))

            self.assertEqual(expected, actual)




#! /usr/bin/env python3

import diagnosticism as d

d.set_program_name('ConRep')

d.report('some important information')
d.report("some slightly less important information: %s, %d, %f" % ('abc', -1, 234.567), show_program_name=False)

d.abort("and we're out")
d.report("ERROR: should not get here!")





#! /usr/bin/env python3

from diagnosticism.logging import (
    enable_logging,
    log,
    set_log_filter,
)
from diagnosticism.program_name import set_program_name
from diagnosticism.severity import *

import unittest
from unittest.mock import patch

import re

try:

    from StringIO import StringIO
except ImportError:

    from io import StringIO


class Logp_tester(unittest.TestCase):


    def test_logging_off(self):

        with patch('sys.stderr', new=StringIO()) as fake_stderr:

            logging_enabled = True if enable_logging(False) else False

            try:

                set_program_name('myprog1')

                log(INFORMATIONAL, 'msg-1')

                self.assertEqual('', fake_stderr.getvalue())
            finally:

                enable_logging(logging_enabled)


    def test_message_as_string(self):

        with patch('sys.stderr', new=StringIO()) as fake_stderr:

            logging_enabled = True if enable_logging(True) else False

            try:

                set_program_name('myprog1')

                log(INFORMATIONAL, 'msg-1')

                self.assertRegex(fake_stderr.getvalue(), r'^\[myprog1, \d+, \d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{6}, .*INFORMATIONAL.*\]: msg-1')
            finally:

                enable_logging(logging_enabled)


    def test__log__WITH_str_AND_file_PARAM(self):

        file = StringIO()

        logging_enabled = True if enable_logging(True) else False

        try:

            set_program_name('myprog3')

            log(INFORMATIONAL, 'msg-3', file=file)
        finally:

            enable_logging(logging_enabled)

        result = file.getvalue()

        self.assertRegex(result, r'^\[myprog3, \d+, \d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{6}, .*INFORMATIONAL.*\]: msg-3')


    def test_message_as_lambda(self):

        with patch('sys.stderr', new=StringIO()) as fake_stderr:

            logging_enabled = True if enable_logging(True) else False

            try:

                set_program_name('myprog1')

                log(INFORMATIONAL, lambda: 'msg-1')

                self.assertRegex(fake_stderr.getvalue(), r'^\[myprog1, \d+, \d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{6}, .*INFORMATIONAL.*\]: msg-1')
            finally:

                enable_logging(logging_enabled)


    def test_threshold_filter(self):

        with patch('sys.stderr', new=StringIO()) as fake_stderr:

            logging = enable_logging(True)

            log_filter = set_log_filter(WARNING)

            try:

                set_program_name('myprog1')

                log(DEBUG0, 'msg-debug0')
                log(INFORMATIONAL, 'msg-informational')
                log(NOTICE, 'msg-notice')
                log(WARNING, 'msg-warning')
                log(FAILURE, 'msg-failure')
                log(CRITICAL, 'msg-critical')
                log(ALERT, 'msg-alert')

                r = fake_stderr.getvalue().rstrip()

                lines = re.split('[\r\n]', r)

                self.assertEqual(4, len(lines))
                self.assertRegex(lines[0], r'^\[myprog1, \d+, \d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{6}, .*WARNING.*\]: msg-warning')
                self.assertRegex(lines[1], r'^\[myprog1, \d+, \d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{6}, .*FAILURE.*\]: msg-failure')
                self.assertRegex(lines[2], r'^\[myprog1, \d+, \d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{6}, .*CRITICAL.*\]: msg-critical')
                self.assertRegex(lines[3], r'^\[myprog1, \d+, \d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{6}, .*ALERT.*\]: msg-alert')
            finally:

                set_log_filter(log_filter[0], log_filter[1])


    def test_dict_filter_others_action_False(self):

        with patch('sys.stderr', new=StringIO()) as fake_stderr:

            logging = enable_logging(True)

            filter_dict = {

                WARNING: True,
                FAILURE: False,
                CRITICAL: True,
                ALERT: True,
                VIOLATION: True,
            }

            log_filter = set_log_filter(filter_dict, False)

            try:

                set_program_name('myprog1')

                log(DEBUG0, 'msg-debug0')
                log(INFORMATIONAL, 'msg-informational')
                log(NOTICE, 'msg-notice')
                log(WARNING, 'msg-warning')
                log(FAILURE, 'msg-failure')
                log(CRITICAL, 'msg-critical')
                log(ALERT, 'msg-alert')

                r = fake_stderr.getvalue().rstrip()

                lines = re.split('[\r\n]', r)

                self.assertEqual(3, len(lines))
                self.assertRegex(lines[0], r'^\[myprog1, \d+, \d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{6}, .*WARNING.*\]: msg-warning')
                self.assertRegex(lines[1], r'^\[myprog1, \d+, \d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{6}, .*CRITICAL.*\]: msg-critical')
                self.assertRegex(lines[2], r'^\[myprog1, \d+, \d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{6}, .*ALERT.*\]: msg-alert')
            finally:

                set_log_filter(log_filter[0], log_filter[1])


    def test_dict_filter_others_action_True(self):

        with patch('sys.stderr', new=StringIO()) as fake_stderr:

            logging = enable_logging(True)

            filter_dict = {

                WARNING: True,
                FAILURE: False,
                CRITICAL: True,
                ALERT: True,
                VIOLATION: True,
            }

            log_filter = set_log_filter(filter_dict, True)

            try:

                set_program_name('myprog1')

                log(DEBUG0, 'msg-debug0')
                log(INFORMATIONAL, 'msg-informational')
                log(NOTICE, 'msg-notice')
                log(WARNING, 'msg-warning')
                log(FAILURE, 'msg-failure')
                log(CRITICAL, 'msg-critical')
                log(ALERT, 'msg-alert')

                r = fake_stderr.getvalue().rstrip()

                lines = re.split('[\r\n]', r)

                self.assertEqual(6, len(lines))
                self.assertRegex(lines[0], r'^\[myprog1, \d+, \d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{6}, .*DEBUG0.*\]: msg-debug0')
                self.assertRegex(lines[1], r'^\[myprog1, \d+, \d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{6}, .*INFORMATIONAL.*\]: msg-informational')
                self.assertRegex(lines[2], r'^\[myprog1, \d+, \d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{6}, .*NOTICE.*\]: msg-notice')
                self.assertRegex(lines[3], r'^\[myprog1, \d+, \d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{6}, .*WARNING.*\]: msg-warning')
                self.assertRegex(lines[4], r'^\[myprog1, \d+, \d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{6}, .*CRITICAL.*\]: msg-critical')
                self.assertRegex(lines[5], r'^\[myprog1, \d+, \d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{6}, .*ALERT.*\]: msg-alert')
            finally:

                set_log_filter(log_filter[0], log_filter[1])


if '__main__' == __name__:

    unittest.main()





from .logging import (
    _do_log,
    _get_log_file_or_default,
)
from . import severity

from .internal import (
    _basename,
    _bool_from_env,
    _is_python_3_9_or_later,
)

import inspect
import sys


_tracingEnabled =   False


def _derive_param(
    param_name, # : str
    params, # : list[str]
):
    """
    xxx.

    Returns
    -------
    tuple
        Tuple of the previous (param_name, param_type, param_value) value(s)
    """

    param_value =   params[param_name]
    param_type  =   param_value.__class__.__name__

    return (param_name, param_type, param_value)


def _dbg(
    file,
    fr,
    show_fileline,
    *args,
    **kwargs,
):

    try:

        fr = fr.f_back

        code = fr.f_code

        file_name   =   code.co_filename
        line_number =   fr.f_lineno

        vnames = code.co_varnames
        params = fr.f_locals

        kwnames = list(kwargs)

        if 0 != len(args):

            s0 = ', '.join(["(%s)=%s" % (type(arg).__name__, arg) for arg in args])
        else:

            s0 = ''

        if 0 != len(kwargs):

            s1 = ', '.join(["%s(%s)=%s" % (name, type(arg).__name__, arg) for name, arg in kwargs.items()])
        else:

            s1 = ''

        if s0 and s1:

            s = ", ".join([s0, s1])
        elif s0:

            s = s0
        elif s1:

            s = s1
        else:

            s = ''

        if show_fileline:

            _log_s(file, severity.TRACE, "%s:%s: %s" % (_basename(file_name), line_number, s))
        else:

            _log_s(file, severity.TRACE, "%s" % s)
    finally:

        del fr


def _flf(
    depth=1,
    **kwargs,
):

    assert 1 == depth

    fr = inspect.currentframe()

    try:

        fr = fr.f_back
        fr = fr.f_back

        code = fr.f_code

        file_name   =   code.co_filename
        line_number =   fr.f_lineno
        if '<module>' == code.co_name:

            function_name   =   "<module>"
        else:

            function_name   =   code.co_name

        if kwargs.get('basename', False):

            file_name = _basename(file_name)

        return [ file_name, line_number, function_name ]
    finally:

        del fr


def _log_s(
    file,
    severity,
    message,
):

    file = _get_log_file_or_default(file)

    _do_log(
        file,
        severity,
        message,
    )


def dbgfl(
    *args,
    **kwargs,
):
    """
    Traces arguments and keyword-arguments with file + line.

    Returns
    -------
    None
    """

    if _tracingEnabled:

        fr = inspect.currentframe()

        file = kwargs.get('file', None)

        _dbg(
            file,
            fr,
            True,
            *args,
            **kwargs,
        )


def dbg(
    *args,
    **kwargs,
):
    """
    Traces arguments and keyword-arguments.

    Returns
    -------
    None
    """

    if _tracingEnabled:

        fr = inspect.currentframe()

        file = kwargs.get('file', None)

        _dbg(
            file,
            fr,
            False,
            *args,
            **kwargs,
        )


def enable_tracing(*args):
    """
    Enables/disables tracing.

    Parameters
    ----------
    *args
        1 or 2 arguments: if 1, then is a `bool` determining whether should be enabled; if 2, then first is name(s) of environment variable(s) to be parsed and second is a `bool` specifying the default if not found in the environment

    Returns
    -------
    bool
        The previous enable/disable setting
    """

    is_enabled = _bool_from_env(args, 'enable_tracing')

    assert(is_enabled == True or is_enabled == False)

    global _tracingEnabled

    _tracingEnabled = is_enabled


def file(
    **kwargs,
):
    """
    Obtains the file of the caller.

    Returns
    -------
    str
    """

    return _flf(depth=1, **kwargs)[0]


def func(
    **kwargs,
):
    """
    Obtains the function of the caller.

    Returns
    -------
    str
    """

    return _flf(depth=1)[2]


def line(
    **kwargs,
):
    """
    Obtains the line of the caller.

    Returns
    -------
    int
    """

    return _flf(depth=1)[1]


def fileline(
    **kwargs,
):
    """
    Obtains a string representing the file and line of the caller.

    Returns
    -------
    str
        String of the form "file:line".
    """

    sep = kwargs.get('sep', ':')

    flf = _flf(depth=1, **kwargs)

    return flf[0] + sep + str(flf[1])


def filelinefunc(
    **kwargs,
):
    """
    Obtains a string representing the file, line, and function of the
    caller.

    Returns
    -------
    str
        String of the form "file:line:func".
    """

    sep = kwargs.get('sep', ':')

    flf = _flf(depth=1, **kwargs)

    return flf[0] + sep + str(flf[1]) + sep + flf[2]


def is_tracing_enabled():
    """
    Indicate whether tracing is enabled

    Returns
    -------
    bool
    """

    return _tracingEnabled


def trace(
    file=None,
):
    """
    Traces function signature, including parameters, of the calling function

    Returns
    -------
    None
    """

    if _tracingEnabled:

        fr = inspect.currentframe()

        try:

            fr = fr.f_back

            code = fr.f_code

            if '<module>' == code.co_name:

                _log_s(file, severity.TRACE, "%s:%d" % (code.co_filename, code.co_firstlineno))
            else:

                fname   =   code.co_name
                vnames  =   code.co_varnames
                params  =   fr.f_locals

                pnames  =   [vname for vname in vnames if vname in params]

                # Algorithm informed by http://blog.mclemon.io/python-efficient-string-concatenation-in-python-2016-edition

                if 0 != len(pnames):

                    pname0 = pnames[0]

                    if 'self' == pname0:

                        val0    =   params[pname0]
                        typ0    =   val0.__class__.__name__

                        plist   =   ", ".join(["%s(%s)=%s" % _derive_param(pname, params) for pname in pnames[1:]])

                        _log_s(file, severity.TRACE, "%s.%s(%s)" % (typ0, fname, plist))

                        return

                plist = ", ".join(["%s(%s)=%s" % _derive_param(pname, params) for pname in pnames])

                _log_s(file, severity.TRACE, "%s(%s)" % (fname, plist))
        finally:

            del fr


if _is_python_3_9_or_later():

    def tracefunc(
        func,
        file=None,
    ):
        """
        Decorator function that equates to the receiver function calling `trace()`

        Returns
        -------
            A wrapper function that calls the decorated function
        """

        def wrapper(*args, **kwargs):

            if is_tracing_enabled():

                sig =   None

                try:

                    sig         =   inspect.signature(func) # py 3.3

                    params      =   sig.parameters
                    params_l    =   list([(k, v) for (k, v) in params.items()])
                    args_n      =   args

                    # S=self, N=normal, A=*args, K=**kwargs

                    paramA      =   None
                    paramK      =   None

                    partN       =   None
                    partA       =   None
                    partK       =   None

                    typeS       =   None

                    if len(params_l) != 0:

                        if 'self' == params_l[0][0]:

                            typeS   =   args[0].__class__.__name__
                            args_n  =   args[1:]

                    if len(params_l) != 0:

                        if params_l[-1][1].kind == inspect.Parameter.VAR_KEYWORD:

                            paramK  =   params_l.pop()
                            partK   =   f"{paramK[0]}(dict)={kwargs}"

                    if len(params_l) != 0:

                        if params_l[-1][1].kind == inspect.Parameter.VAR_POSITIONAL:

                            paramA  =   params_l.pop()
                            partA   =   f"{paramA[0]}(tuple)={args_n[len(params_l):]}"

                    if len(params_l) != 0:

                        namesN  =   [p[0] for p in params_l]
                        partN   =   ', '.join([f"{k}({v.__class__.__name__})={v}" for (k, v) in dict(zip(namesN, args_n)).items()])


                    fname   =   func.__name__


                    args_list = []

                    if partN and len(partN) != 0:

                        args_list.append(partN)

                    if partA and len(partA) != 0:

                        args_list.append(partA)

                    if partK and len(partK) != 0:

                        args_list.append(partK)

                    plist = ', '.join(args_list)


                    if typeS:

                        _log_s(file, severity.TRACE, "%s.%s(%s)" % (typeS, fname, plist))
                    else:

                        _log_s(file, severity.TRACE, "%s(%s)" % (fname, plist))
                finally:

                    del sig

            return func(*args, **kwargs)

        return wrapper

    def asynctracefunc(
        func,
        file=None,
    ):
        """
        Async decorator function that equates to the receiver function calling `trace()`

        Returns
        -------
            A wrapper function that calls the decorated function
        """

        return tracefunc(func, file)


# ############################## end of file ############################# #




#! /usr/bin/env python3

import asyncio
import diagnosticism as d

@d.asynctracefunc
async def func1():

    pass


@d.asynctracefunc
async def func2(x, y):

    pass


@d.asynctracefunc
async def func3(x, y, **options):

    z = x * y

    return z


@d.asynctracefunc
async def func4(x, y, *args):

    pass


@d.asynctracefunc
async def func5(x, y, *args, **options):

    pass


class Thing:


    @d.tracefunc
    def __init__(self):

        pass


    @d.asynctracefunc
    async def some_method(self, x, y, z):

        pass


@d.asynctracefunc
async def main():

    await asyncio.create_task(func1())

    await asyncio.create_task(func2('abc', 12))

    await asyncio.create_task(func3('abc', 12))
    await asyncio.create_task(func3('abc', 12, pos='left'))

    await asyncio.create_task(func4('abc', 12))
    await asyncio.create_task(func4('abc', 12, 13, 14, 15))

    await asyncio.create_task(func5('abc', 12))
    await asyncio.create_task(func5('abc', 12, 13, 14, 15))
    await asyncio.create_task(func5('abc', 12, pos='left', length=10))
    await asyncio.create_task(func5('abc', 12, 13, 14, 15, pos='left', length=10))

    thing = Thing()
    await asyncio.create_task(thing.some_method(-1, 0, +1))

    await asyncio.sleep(2.0)


d.enable_tracing(('DIAGNOSTICISM_ENABLE_TRACING', 'ENABLE_TRACING'), True)

asyncio.run(main())




from .contingent_reporting import (
    _do_report,
    _get_cr_file_or_default,
    _isatty,
)
from .program_name import get_program_name
from . import severity as _severity

from .internal import (
    _bool_from_env,
)

from datetime import datetime as dt
import sys
import threading


_REAL_STDERR = sys.stderr

def _get_stderr_dynamic():

    return sys.stderr

def _get_stderr_static():

    return _REAL_STDERR


def _get_log_file_or_default(file):

    if file:

        return file

    return _get_stderr_dynamic()


class _clrs:

    RED         =   "\x1B[31m"
    GREEN       =   "\x1B[32m"
    YELLOW      =   "\x1B[33m"
    BLUE        =   "\x1B[34m"
    MAGENTA     =   "\x1B[35m"
    DARKGREY    =   "\x1B[90m"
    NONE        =   "\x1B[0m"

    D   =   {

        _severity.VIOLATION         :   RED,
        _severity.ALERT             :   RED,
        _severity.CRITICAL          :   RED,
        _severity.FAILURE           :   RED,
        _severity.WARNING           :   YELLOW,
        _severity.NOTICE            :   MAGENTA,
        _severity.INFORMATIONAL     :   MAGENTA,
        _severity.DEBUG0            :   BLUE,
        _severity.DEBUG1            :   BLUE,
        _severity.DEBUG2            :   BLUE,
        _severity.DEBUG3            :   BLUE,
        _severity.DEBUG4            :   BLUE,
        _severity.DEBUG5            :   BLUE,
        _severity.TRACE             :   DARKGREY,
        _severity.BENCHMARK         :   DARKGREY,
    }


# ######################################
# Runtime shared state

_logging_is_enabled =   None
_log_filter         =   None
_others_action      =   None


def enable_logging(*args):
    """
    Enables/disables logging.

    Parameters
    ----------
    *args
        1 or 2 arguments: if 1, then is a `bool` determining whether should be enabled; if 2, then first is name(s) of environment variable(s) to be parsed and second is a `bool` specifying the default if not found in the environment

    Returns
    -------
    bool
        The previous enable/disable setting

    Raises
    ------
    ValueError
        If the string form of `v` does not contain a recognisable severity level
    """

    is_enabled = _bool_from_env(args, 'enable_logging')

    assert(is_enabled == True or is_enabled == False or is_enabled is None)

    global _logging_is_enabled

    _logging_is_enabled, is_enabled = is_enabled, _logging_is_enabled

    return is_enabled


def is_logging_enabled():
    """
    Indicate whether logging is enabled
    """

    return _logging_is_enabled


def is_severity_logged(severity):
    """
    Indicates whether the given severity is logged

    Returns
    -------
    bool
        `True` if the severity is to be logged; `False` otherwise
    """

    if not _logging_is_enabled:

        return False

    if _log_filter:

        if isinstance(_log_filter, (dict, )):

            r = _log_filter.get(severity, _others_action)

            if not r:

                return False
        else:

            if int(severity) > int(_log_filter):

                return False

    return True


def set_log_filter(
    log_filter,
    others_action=None,
):
    """
    Sets a logging filter, which may either specify a threshold severity or a mapping of levels to actions

    Parameters
    ----------
    log_filter : int, dict
        If a dictionary, it is interpreted as a mapping from severity-level to `True`/`False` that controls each level's output; otherwise, treated as a severity-level threshold (and must be convertible to `int`)

    others_action : obj, optional
        An object to be passed to be used in the case that the log_filter is a dictionary and the severity is not recognised

    Returns
    -------
    tuple
        Tuple of the previous (log_filter, others_action) value(s)
    """

    assert log_filter is None or isinstance(log_filter, (dict, )) or isinstance(int(log_filter), int)

    global _log_filter, _others_action

    _log_filter, log_filter, _others_action, others_action = log_filter, _log_filter, others_action, _others_action

    return (log_filter, others_action)


def _do_log(
    file,
    severity,
    message,
):
    """
    INTERNAL FUNCTION ONLY
    """

    assert file

    now = dt.now()

    ss = _severity.severity_to_string(severity)

    if _isatty(file):

        r = _clrs.D.get(severity, None)

        if r:

            ss = r + ss + _clrs.NONE

    # TODO: perf test this
    prefix = "[%s, %s, %s, %s]" % (get_program_name(), threading.current_thread().ident, str(now), ss)

    if hasattr(message, '__call__'):
        message = message()

    # TODO: perf test this
    full = prefix + ': ' + message

    show_program_name = False
    trailing_prompt = False

    file = _get_cr_file_or_default(file)

    _do_report(
        file,
        full,
        show_program_name,
        trailing_prompt,
    )


def log(
    severity,
    message,
    file=None,
):
    """
    Conditionally issue the given log message based on the given severity

    Parameters
    ----------
    severity : int
        The severity associated with the message

    message : str, callable
        A message string to be emitted, or a callable object that will yield an emittable string when called

    Returns
    -------
    None
    """

    if not is_severity_logged(severity):

        return

    file = _get_log_file_or_default(file)

    return _do_log(
        file,
        severity,
        message,
    )


# ############################## end of file ############################# #





from .program_name import get_program_name

from .internal import (
    _is_python_2_7_or_later,
    _is_windows_11_or_later,
)

import os
import sys


def _supports_ansi_sequences():

    if os.name.lower() == 'posix':

        return True

    if _is_windows_11_or_later():

        return True

    return False


_OS_IS_POSIX = _supports_ansi_sequences()

STOCK_TRAILING_PROMPT = 'use --help for usage'
"""The trailing prompt used when trailing_prompt=True is passed to abort() (and family) and no default trailing prompt has been set (via set_default_trailing_prompt())"""

_trailing_prompt = None


def set_default_trailing_prompt(prompt):
    """
    Sets the default trailing prompt.

    Parameters
    ----------
    prompt : str
        The new default trailing prompt

    Returns
    -------
    str
        The previous default trailing prompt

    Notes
    -----
        This method is NOT threadsafe, so in a multithreaded context the set prompt and/or the return value may not be as expected
    """

    global _trailing_prompt

    _trailing_prompt, prompt = prompt, _trailing_prompt

    return prompt


def _emit_to_cr_stm(
    file,
    message,
):
    assert file

    file.write(message)


def _get_cr_file_or_default(file):

    if file:

        return file

    return sys.stderr


def _add_eol_and_emit_to_cr_stm(
    file,
    message,
):

    _emit_to_cr_stm(
        file,
        message + "\n",
    )


def _isatty(
    file,
):

    if not _OS_IS_POSIX:

        return False

    file = _get_cr_file_or_default(file)

    if not file.isatty():

        return False

    return True


def _do_report(
    file,
    message,
    program_name,
    trailing_prompt,
):
    assert file

    # first process trailing prompt (as more complex)

    if False != trailing_prompt:

        if False:

            pass
        elif True == trailing_prompt:

            dtp = _trailing_prompt

            if dtp:

                trailing_prompt = dtp
            else:

                trailing_prompt = STOCK_TRAILING_PROMPT
        elif None == trailing_prompt:

            dtp = _trailing_prompt

            if dtp:

                trailing_prompt = dtp
        else:

            trailing_prompt = str(trailing_prompt)

            #trailing_prompt = trailing_prompt.strip()

            if not trailing_prompt:

                trailing_prompt = False

    if program_name:

        if True == program_name:

            program_name = get_program_name()

    if program_name:

        if trailing_prompt:

            # TODO: perf test this
            _emit_to_cr_stm(file, program_name + ": " + str(message) + '; ' + trailing_prompt + "\n")
        else:

            # TODO: perf test this
            _emit_to_cr_stm(file, program_name + ": " + str(message) + "\n")
    else:

        if trailing_prompt:

            # TODO: perf test this
            _emit_to_cr_stm(file, message + '; ' + trailing_prompt + "\n")
        else:

            _add_eol_and_emit_to_cr_stm(file, message)


if _is_python_2_7_or_later():

    def conrep(
        message,
        **kwargs,
    ):
        """
        DEPRECATED: use `report()`.
        """

        show_program_name = kwargs.get('show_program_name', True)

        return report(
            message,
            show_program_name=show_program_name,
        )


def report(
    message,
    show_program_name=True,
    file=None,
):
    """
    Emits the given message (and optional prefix) on the contingent report stream.

    Parameters
    ----------
    message : str
        The (main body of the) message to be displayed, which is optionally prefixed by program-name + ': '

    show_program_name : bool, optional
        Prevents the default prefixing of the message with program-name + ': ' if `False`. Default is `True`
    """

    trailing_prompt = False

    file = _get_cr_file_or_default(file)

    _do_report(
        file,
        message,
        show_program_name,
        trailing_prompt,
    )


def abort(
    message,
    do_exit=True,
    show_program_name=True,
    trailing_prompt=None,
    file=None,
):
    """
    Emits the given message (and optional prefix and suffix) on the contingent report stream, and then terminates the process

    Parameters
    ----------
    message : str
        The (main body of the) message to be displayed, which is optionally prefixed by program-name + ': ' and optionally suffixed by '; ' + trailing-prompt

    do_exit : bool, optional
        Prevents default behaviour of a call to `sys.exit(1)` (after printing the abort message) if `False`. Default is `True`

    show_program_name : bool, optional
        Prevents the default prefixing of the message with program-name + ': ' if `False`. Default is `True`

    trailing_prompt : str, bool, optional
        Affects the use of the trailing prompt as follows: if `False`, no trailing prompt is shown; if a (non-empty) string, that is used; if `None`, the default trailing prompt, if any, is used; if `True`, the default trailing prompt is used if specified, otherwise `STOCK_TRAILING_PROMPT` is used
    """

    file = _get_cr_file_or_default(file)

    _do_report(
        file,
        message,
        show_program_name,
        trailing_prompt,
    )

    if do_exit:

        sys.exit(1)


# ############################## end of file ############################# #




#! /usr/bin/env python3

import diagnosticism as d
import diagnosticism.severity as sev

import os


ENV_VAR_NAME = 'DP_LOG_THRESHOLD'
FILTER_LEVEL = sev.INFO


d.enable_logging(('DIAGNOSTICISM_ENABLE_LOGGING', 'ENABLE_LOGGING'), True)


if ENV_VAR_NAME in os.environ:

    try:

        threshold = int(os.environ[ENV_VAR_NAME])

        d.set_log_filter(threshold)
    except ValueError:

        d.abort("environment variable '%s' exists and has a value not convertible to `int`" % (ENV_VAR_NAME))
else:

    d.set_log_filter(FILTER_LEVEL)

d.log(sev.BENCHMARK, 'benchmark')
d.log(sev.TRACE, 'trace')
d.log(sev.DEBUG5, 'debug5')
d.log(sev.DEBUG4, 'debug4')
d.log(sev.DEBUG3, 'debug3')
d.log(sev.DEBUG2, 'debug2')
d.log(sev.DEBUG1, 'debug1')
d.log(sev.DEBUG0, 'debug0')
d.log(sev.INFO, 'informational')
d.log(sev.NOTICE, 'notice')
d.log(sev.WARNING, 'warning')
d.log(sev.FAILURE, 'failure')
d.log(sev.CRITICAL, 'critical')
d.log(sev.ALERT, "alert")
d.log(sev.VIOLATION, "VIOLATION")





import os
import sys


_program_name = None


def set_program_name(name):
    """
    Sets the program name

    Parameters
    ----------
    name : str
        The string to be used for the program name

    Returns
    -------
    str
        The previous program name
    """

    global _program_name

    if name:

        _program_name = name
    else:

        _program_name = os.path.basename(sys.argv[0])

    return _program_name


def get_program_name():
    """
    Gets the program name
    """

    return _program_name


# initialise the program name
set_program_name(None)


# ############################## end of file ############################# #




#! /usr/bin/env python3

from diagnosticism import (
    dbg,
    dbgfl,
    enable_tracing,
)

import sys


def f():

    print("about to call dbg on an int and a string")

    dbg()
    dbg(10)
    dbg(10, 20)
    dbg(i=10)
    dbg(i=10, j='abc')

    dbg('s', 't')


    dbgfl()
    dbgfl(10)
    dbgfl(10, 20)
    dbgfl(i=10)
    dbgfl(i=10, j='abc')

    dbgfl('s', 't')


if __name__ == "__main__":

    enable_tracing(True)

    print("")
    print("`main()`:")

    print("about to call dbg on an int and a string")

    dbg()
    dbg(10)
    dbg(10, 20)
    dbg(i=10)
    dbg(i=10, j='abc')

    dbg('s')
    dbg('s', 't')


    dbgfl()
    dbgfl(10)
    dbgfl(10, 20)
    dbgfl(i=10)
    dbgfl(i=10, j='abc')

    dbgfl('s', 't')



    print("")
    print("`f()`:")

    f()





from .contingent_reporting import (
    _add_eol_and_emit_to_cr_stm,
    _get_cr_file_or_default,
)
from .logging import (
    _do_log,
    _get_log_file_or_default,
    is_logging_enabled,
)
from . import severity

import sys


def _warn(
    file_cr,
    file_dl,
    message_lines,
):

    file_cr = _get_cr_file_or_default(file_cr)

    if is_logging_enabled():

        file_dl = _get_log_file_or_default(file_dl)

    for message in message_lines:

        if message:

            if is_logging_enabled():

                _do_log(
                    file_dl,
                    severity.WARN,
                    message,
                )

            _add_eol_and_emit_to_cr_stm(
                file_cr,
                message,
            )


def warn(
    *message_lines,
    file=None,
    file_cr=None,
    file_dl=None,
):
    """
    Analogue of Ruby's `Kernel#warn()`

    Parameters
    ----------
    message : str, None
        The message to be emitted to the standard error stream (along with a new-line sequence). If `None`, nothing is emitted

    file : file-object, optional
        An object with a `write(str)` method, or `None` (or not present), in which case the default of `sys.stderr` will be used

    file_cr : file-object, optional
        An object with a `write(str)` method, or `None` (or not present), in which case the default given in `file` will be used. Used for contingent report output

    file_dl : file-object, optional
        An object with a `write(str)` method, or `None` (or not present), in which case the default given in `file` will be used. Used for diagnostic logging output

    Returns
    -------
    None
    """

    if file_cr is None:

        file_cr = file

    if file_dl is None:

        file_dl = file

    _warn(
        file_cr,
        file_dl,
        message_lines,
    )


# ############################## end of file ############################# #





import os
import platform
import sys


TRUE_STRINGS    =   [
    'true',
    'True',
    'TRUE',
    'on',
    'On',
    'ON',
    'yes',
    'Yes',
    'YES',
    '1',
]

FALSE_STRINGS   =   [
    'false',
    'False',
    'FALSE',
    'off',
    'Off',
    'OFF',
    'no',
    'No',
    'NO',
    '0',
]

TRUE_STRINGS_lower  =   [
    'true',
    'on',
    'yes',
    '1',
]

FALSE_STRINGS_lower =   [
    'false',
    'off',
    'no',
    '0',
]


def _is_windows():
    """
    Obtains whether the operating system is a flavour of Windows.

    Returns
    -------
    bool
    """

    if os.name.lower() == 'nt':

        return True

    return False


def _is_windows_11_or_later():
    """
    Obtains whether the operating system is a flavour of Windows that is
    11 or later.

    Returns
    -------
    bool
    """

    if not _is_windows():

        return False

    from platform import release as platform_release
    from platform import version as platform_version

    release = platform_release()

    try:

        release_i = int(release)

        if release_i < 10:

            return False
    except ValueError:

        return False

    version = platform_version().split('.')

    if len(version) < 3:

        return False

    try:

        version_major   =   int(version[0])
        version_minor   =   int(version[1])
        version_patch   =   int(version[2])

        if version_major > 10:

            return True

        if version_major < 10:

            return False

        if version_minor > 0:

            return True

        if version_patch >= 22000:

            return True

    except ValueError:

        return False


def _basename(path):
    """
    Obtains the basename form of the path.

    Returns
    -------
    str
    """

    ix_last_slash   =   path.rfind('/')

    if _is_windows():

        ix_last_bslash  =   path.rfind('\\')

        if ix_last_bslash > ix_last_slash:

            ix_last_slash = ix_last_bslash

    return path[ix_last_slash + 1:]


def _str2bool(s, default_value=None):
    """
    Converts a string to a boolean, by consideration of the human-readable
    intention of the string. For example,

    Returns
    -------
    True | False | None
    """

    if s in TRUE_STRINGS:

        return True


    if s in FALSE_STRINGS:

        return False


    s = s.trim().lower()


    if s in TRUE_STRINGS_lower:

        return True


    if s in FALSE_STRINGS_lower:

        return False


    return default_value


def _bool_from_env(args, fn_name):

    if len(args) == 1:

        v = args[0]

        if not isinstance(v, bool):

            raise TypeError("`%s()` argument must be `True` or `False`; type `%s` provided" % (fn_name, type(v)))

        return v

    if len(args) == 2:

        def_val =   args[1]

        if isinstance(args[0], (list, tuple)):

            ev_names    =   args[0]
        else:

            ev_names    =   [ args[0] ]

        for ev_name in ev_names:

            if ev_name in os.environ:

                ev_val = os.environ[ev_name]

                return _str2bool(ev_val, def_val)

        if not isinstance(def_val, bool):

            raise TypeError("`%s()` argument must be `True` or `False`; type `%s` provided" % (fn_name, type(def_val)))

        return def_val

    raise TypeError("`%s()` takes 1 or 2 arguments" % fn_name)


def _is_python_2_7_or_later():

    return sys.version_info[:2] >= (2, 7)


def _is_python_3_0_or_later():

    return sys.version_info[:2] >= (3, 0)


def _is_python_3_9_or_later():

    return sys.version_info[:2] >= (3, 9)


# ############################## end of file ############################# #




#! /usr/bin/env python3

# ######################################################################## #
# File:     tests/test_warn.py
#
# Purpose:  Unit-testing of `warn()`.
#
# Created:  25th July 2025
# Updated:  25th July 2025
#
# Copyright (c) Matthew Wilson, Synesis Information Systems Pty Ltd
# All rights reserved
#
# ######################################################################## #


from diagnosticism.logging import (
    enable_logging,
)
from diagnosticism.warning import (
    warn,
)
from diagnosticism.program_name import (
    set_program_name,
)

import re
import unittest
from unittest.mock import patch

try:

    from StringIO import StringIO
except ImportError:

    from io import StringIO


class Warn_tester(unittest.TestCase):


    def test__warn__WITH_PROGRAM_NAME_AND_LOGGING_DISABLED(self):

        with patch('sys.stderr', new=StringIO()) as fake_stderr:

            set_program_name('myprog1')

            logging_enabled = True if enable_logging(False) else False

            try:

                warn('warning-1')
            finally:

                enable_logging(logging_enabled)

            result = fake_stderr.getvalue()

            self.assertEqual("warning-1\n", result)


    def test__warn__WITH_PROGRAM_NAME_AND_LOGGING_ENSABLED(self):

        with patch('sys.stderr', new=StringIO()) as fake_stderr:

            set_program_name('myprog2')

            logging_enabled = True if enable_logging(True) else False

            try:

                warn('warning-2')
            finally:

                enable_logging(logging_enabled)

            result = fake_stderr.getvalue()

            regex = r"^\[myprog2,\s*\d+,\s*\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}.\d{6}, WARNING]: warning-2\nwarning-2\n$"

            self.assertRegex(result, regex)


    def test__warn__WITH_file_PARAM_AND_LOGGING_DISABLED(self):

        file = StringIO()

        set_program_name('myprog3')

        logging_enabled = True if enable_logging(False) else False

        try:

            warn('warning-3', file=file)
        finally:

            enable_logging(logging_enabled)

        result = file.getvalue()

        self.assertEqual("warning-3\n", result)


    def test__warn__WITH_file_PARAM_AND_LOGGING_ENABLED(self):

        file = StringIO()

        set_program_name('myprog4')

        logging_enabled = True if enable_logging(True) else False

        try:

            warn('warning-4', file=file)
        finally:

            enable_logging(logging_enabled)

        result = file.getvalue()

        regex = r"^\[myprog4,\s*\d+,\s*\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}.\d{6}, WARNING]: warning-4\nwarning-4\n$"

        self.assertRegex(result, regex)


    def test__warn__WITH_file_cr_PARAM_AND_file_dl_PARAM_AND_LOGGING_ENABLED(self):

        file_cr = StringIO()
        file_dl = StringIO()

        set_program_name('myprog4')

        logging_enabled = True if enable_logging(True) else False

        try:

            warn('warning-4', file_cr=file_cr, file_dl=file_dl)
        finally:

            enable_logging(logging_enabled)

        result_cr = file_cr.getvalue()
        result_dl = file_dl.getvalue()

        regex = r"^\[myprog4,\s*\d+,\s*\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}.\d{6}, WARNING]: warning-4\n$"

        self.assertEqual("warning-4\n", result_cr)
        self.assertRegex(result_dl, regex)



    def test__warn__WITH_file_PARAM_AND_file_dl_PARAM_AND_LOGGING_ENABLED(self):

        file = StringIO()
        file_dl = StringIO()

        set_program_name('myprog4')

        logging_enabled = True if enable_logging(True) else False

        try:

            warn('warning-4', file=file, file_dl=file_dl)
        finally:

            enable_logging(logging_enabled)

        result = file.getvalue()
        result_dl = file_dl.getvalue()

        regex = r"^\[myprog4,\s*\d+,\s*\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}.\d{6}, WARNING]: warning-4\n$"

        self.assertEqual("warning-4\n", result)
        self.assertRegex(result_dl, regex)


    def test__warn__WITH_MULTIPLE_LINES_WITH_PROGRAM_NAME_AND_LOGGING_DISABLED(self):

        with patch('sys.stderr', new=StringIO()) as fake_stderr:

            set_program_name('myprog5')

            logging_enabled = True if enable_logging(False) else False

            try:

                warn(
                    'warning-5a',
                    'warning-5b',
                )
            finally:

                enable_logging(logging_enabled)

            result = fake_stderr.getvalue()

            self.assertEqual("warning-5a\nwarning-5b\n", result)




#! /usr/bin/env python3

import diagnosticism as d

d.set_program_name('ConRep')

d.abort("abort with application-defined TRAILING_PROMPT", trailing_prompt='use --help for help', do_exit=False)

d.abort("default abort", do_exit=False)

d.abort("default abort with default TRAILING_PROMPT", do_exit=False, trailing_prompt=True)

print("setting default trailing prompt")
d.contingent_reporting.set_default_trailing_prompt('use --help to show usage')

d.abort("default abort with default TRAILING_PROMPT", do_exit=False, trailing_prompt=True)

d.abort("default abort", do_exit=False)





#! /usr/bin/env python3

from diagnosticism.contingent_reporting import abort as d_abort
from diagnosticism.program_name import get_program_name
from diagnosticism.severity import parse_verbosity

import sys


if len(sys.argv) != 2:

    d_abort("wrong number of arguments", trailing_prompt=True)
else:

    if '--help' == sys.argv[1]:

        print("USAGE: %s <severity-name-or-value>" % (get_program_name()))

        sys.exit(0)
    else:

        severity_string = sys.argv[1]


try:

    severity = parse_verbosity(severity_string)

    print("given severity string '%s' is parsed as %d" % (severity_string, severity))
    # d_abort(

except ValueError as x:

    print("exception (%s): %s" % (type(x), str(x)), file=sys.stderr, flush=True)





# ######################################################################## #
# File:     doomgram.py
#
# Purpose:  Definition of the `DOOMGram` class.
#
# Created:  19th July 2025
# Updated:  24th July 2025
#
# Author:   Matthew Wilson
#
# Copyright (c) 2025, Matthew Wilson and Synesis Information Systems
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# ######################################################################## #


import math
import time


class DOOMGram:
    """
    Decimal Order-Of-Magnitude frequency histoGRAM.
    """

    __slots__ = (

        '_event_count',
        '_total_event_time_ns',
        '_min_event_time_ns',
        '_max_event_time_ns',
        '_num_events_in_1ns',
        '_num_events_in_10ns',
        '_num_events_in_100ns',
        '_num_events_in_1us',
        '_num_events_in_10us',
        '_num_events_in_100us',
        '_num_events_in_1ms',
        '_num_events_in_10ms',
        '_num_events_in_100ms',
        '_num_events_in_1s',
        '_num_events_in_10s',
        '_num_events_ge_100s',
    )

    def __init__(self):

        self._event_count           =   0

        self._total_event_time_ns   =   0

        self._min_event_time_ns     =   None
        self._max_event_time_ns     =   None

        self._num_events_in_1ns     =   0
        self._num_events_in_10ns    =   0
        self._num_events_in_100ns   =   0
        self._num_events_in_1us     =   0
        self._num_events_in_10us    =   0
        self._num_events_in_100us   =   0
        self._num_events_in_1ms     =   0
        self._num_events_in_10ms    =   0
        self._num_events_in_100ms   =   0
        self._num_events_in_1s      =   0
        self._num_events_in_10s     =   0
        self._num_events_ge_100s    =   0

    def event_count(self):
        """
        Number of events counted.
        """

        return self._event_count

    def total_event_time_ns(self):
        """
        Obtains the total event time (in nanoseconds).
        """

        return self._total_event_time_ns

    def min_event_time_ns(self):
        """
        Obtain the minimum event time (in nanoseconds), or +None+ if no
        events.
        """

        return self._min_event_time_ns

    def max_event_time_ns(self):
        """
        Obtain the maximum event time (in nanoseconds), or +None+ if no
        events.
        """

        return self._max_event_time_ns

    def num_events_in_1ns(self):
        """
        Number of events counted in the interval [1ns, 10ns).
        """

        return self._num_events_in_1ns

    def num_events_in_10ns(self):
        """
        Number of events counted in the interval [10ns, 100ns).
        """

        return self._num_events_in_10ns

    def num_events_in_100ns(self):
        """
        Number of events counted in the interval [100ns, 1µs).
        """

        return self._num_events_in_100ns

    def num_events_in_1us(self):
        """
        Number of events counted in the interval [1µs, 10µs).
        """

        return self._num_events_in_1us

    def num_events_in_10us(self):
        """
        Number of events counted in the interval [10µs, 100µs).
        """

        return self._num_events_in_10us

    def num_events_in_100us(self):
        """
        Number of events counted in the interval [100µs, 1ms).
        """

        return self._num_events_in_100us

    def num_events_in_1ms(self):
        """
        Number of events counted in the interval [1ms, 10ms).
        """

        return self._num_events_in_1ms

    def num_events_in_10ms(self):
        """
        Number of events counted in the interval [10ms, 100ms).
        """

        return self._num_events_in_10ms

    def num_events_in_100ms(self):
        """
        Number of events counted in the interval [100ms, 1s).
        """

        return self._num_events_in_100ms

    def num_events_in_1s(self):
        """
        Number of events counted in the interval [1s, 10s).
        """

        return self._num_events_in_1s

    def num_events_in_10s(self):
        """
        Number of events counted in the interval [10s, 100s).
        """

        return self._num_events_in_10s

    def num_events_ge_100s(self):
        """
        Number of events counted in the interval [100s, ∞).
        """

        return self._num_events_ge_100s

    def push_event_time_ns(self, time_in_ns):
        """
        Pushes an event with the given number of nanoseconds.
        """

        assert isinstance(time_in_ns, int)

        self._event_count += 1

        self._total_event_time_ns += time_in_ns


        if self._min_event_time_ns is None or time_in_ns < self._min_event_time_ns:

            self._min_event_time_ns = time_in_ns


        if self._max_event_time_ns is None or self._max_event_time_ns < time_in_ns:

            self._max_event_time_ns = time_in_ns



        if time_in_ns >= 100_000_000:

            if time_in_ns >= 10_000_000_000:

                if time_in_ns >= 100_000_000_000:

                    self._num_events_ge_100s += 1
                else:

                    self._num_events_in_10s += 1

            else:

                if time_in_ns >= 1_000_000_000:

                    self._num_events_in_1s += 1
                else:

                    self._num_events_in_100ms += 1

        else:

            if time_in_ns >= 10_000:

                if time_in_ns >= 1_000_000:

                    if time_in_ns >= 10_000_000:

                        self._num_events_in_10ms += 1
                    else:

                        self._num_events_in_1ms += 1

                else:

                    if time_in_ns >= 100_000:

                        self._num_events_in_100us += 1
                    else:

                        self._num_events_in_10us += 1
            else:

                if time_in_ns >= 100:

                    if time_in_ns >= 1_000:

                        self._num_events_in_1us += 1
                    else:

                        self._num_events_in_100ns += 1
                else:

                    if time_in_ns >= 10:

                        self._num_events_in_10ns += 1
                    elif time_in_ns >= 1:

                        self._num_events_in_1ns += 1

    def push_event_time_us(self, time_in_us):
        """
        Pushes an event with the given number of microseconds.
        """

        self.push_event_time_ns(time_in_us * 1_000)

    def push_event_time_ms(self, time_in_ms):
        """
        Pushes an event with the given number of milliseconds.
        """

        self.push_event_time_ns(time_in_ms * 1_000_000)

    def push_event_time_s(self, time_in_s):
        """
        Pushes an event with the given number of seconds.
        """

        self.push_event_time_ns(time_in_s * 1_000_000_000)

    def to_strip(self, **kwargs):
        """
        Converts to string form according to given options

        === Signature

        * *Parameters:*
          - +options+ (+Hash+, +Integer+) Combination of flags (with behaviour as described below for the +flags+ option), or an options hash;

        * *Options:*
          - +overflow_character+ (+String+) A string (of length 1) that specifies the symbol for counts outside the available range. Defaults to +'*'+;
          - +range+ (+String+) A string whose characters specificy the symbols to use for counts in orders of magnitude. Defaults to +'abcdefghijklmnopqrstuvwxyz'+, which caters to the counts 1-9 => +'a', 10-99 => +'b'+, 100-999 => +'c'+, ... 10^25-(10^26-1) => +'z'+;
          - +zero_character+ (+String+) A string (of length 1) that specifies the symbol for a count of 0. Defaults to +' '+;

        === Return
        (+String+) A string (of length 12) containing symbols representing the counts in the ranges 1ns, 10ns, ..., 10s, 100+s.
        """

        ch_zero     =   kwargs.get('zero', '_')
        ch_overflow =   kwargs.get('overflow', '*')
        range_chars =   kwargs.get('range', 'abcdefghijklmnopqrstuvwxyz')

        counts = [
            self._num_events_in_1ns,
            self._num_events_in_10ns,
            self._num_events_in_100ns,
            self._num_events_in_1us,
            self._num_events_in_10us,
            self._num_events_in_100us,
            self._num_events_in_1ms,
            self._num_events_in_10ms,
            self._num_events_in_100ms,
            self._num_events_in_1s,
            self._num_events_in_10s,
            self._num_events_ge_100s,
        ]

        s = [ch_zero[0]] * 12 # "____________"

        for index, count in enumerate(counts):

            if 0 == count:

                continue

            # TODO: get a faster way to count digits

            oom = int(math.log10(count))

            if oom < len(range_chars):

                s[index] = range_chars[oom]
            else:

                s[index] = ch_overflow[0]

        return ''.join(s)

    def __repr__(self):
        """
        A representation of the `DOOMGram` instance that shows all internal
        state.
        """

        return "%s(%s=%d, %s=%d, %s=%s, %s=%s, %s=%d, %s=%d, %s=%d, %s=%d, %s=%d, %s=%d, %s=%d, %s=%d, %s=%d, %s=%d, %s=%d, %s=%d)" % (
            'DOOMGram',
            'event_count',          self.event_count(),
            'total_event_time_ns',  self.total_event_time_ns(),
            'min_event_time_ns',    self.min_event_time_ns(),
            'max_event_time_ns',    self.max_event_time_ns(),
            'num_events_in_1ns',    self.num_events_in_1ns(),
            'num_events_in_10ns',   self.num_events_in_10ns(),
            'num_events_in_100ns',  self.num_events_in_100ns(),
            'num_events_in_1us',    self.num_events_in_1us(),
            'num_events_in_10us',   self.num_events_in_10us(),
            'num_events_in_100us',  self.num_events_in_100us(),
            'num_events_in_1ms',    self.num_events_in_1ms(),
            'num_events_in_10ms',   self.num_events_in_10ms(),
            'num_events_in_100ms',  self.num_events_in_100ms(),
            'num_events_in_1s',     self.num_events_in_1s(),
            'num_events_in_10s',    self.num_events_in_10s(),
            'num_events_ge_100s',   self.num_events_ge_100s(),
        )

    def __str__(self):
        """
        Converts to string form according to default options.
        """

        return self.to_strip()


class DOOMScope:
    """
    Scoping class for histogram timings using `DOOMScope`.
    """

    __slots__ = (
        '_dg',
        '_before',
    )

    def __init__(self, dg):

        assert isinstance(dg, DOOMGram)

        self._dg = dg
        self._before = None

    def __enter__(self):

        self._before = time.perf_counter_ns()

        return self._dg

    def __exit__(self, x_type, x_val, x_tb):

        after = time.perf_counter_ns()

        self._dg.push_event_time_ns(after - self._before)


# ############################## end of file ############################# #





__author__      =   'Matt Wilson'
__copyright__   =   'Copyright 2019-2025, Synesis Information Systems, Copyright 2019, Synesis Software'
__credits__     =   [
    'Garth Lancaster',
    'Matt Wilson',
]
__email__       =   'matthew@synesis.com.au'
__license__     =   'BSD-3-Clause'
__maintainer__  =   'Matt Wilson'
__status__      =   'Beta'
__version__     =   '0.15.0'

from .contingent_reporting import (
    abort,
    report,
)
from .doomgram import (
    DOOMGram,
    DOOMScope,
)
from .logging import (
    enable_logging,
    is_logging_enabled,
    is_severity_logged,
    log,
    set_log_filter,
)
from .program_name import *
from .severity import *
from .tracing import (
    dbg,
    dbgfl,
    enable_tracing,
    file,
    fileline,
    filelinefunc,
    func,
    is_tracing_enabled,
    line,
    trace,
    tracefunc,
    asynctracefunc,
)
from .warning import warn


# ############################## end of file ############################# #




#! /usr/bin/env python3

from diagnosticism.program_name import set_program_name
from diagnosticism.tracing import (
    enable_tracing,
    trace,
    tracefunc,
)

import unittest
from unittest.mock import patch

import re

try:

    from StringIO import StringIO
except ImportError:

    from io import StringIO

file = None


def f1():

    trace(file=file)

    return "f1-result"

@tracefunc
def f2():

    return "f2-result"


class Trace_tester(unittest.TestCase):


    def test__trace__WITH_TRACING_DISABLED(self):

        tracing_enabled = True if enable_tracing(False) else False

        try:

            set_program_name('myprog1')

            global file

            file = StringIO()

            try:

                r = f1()

                self.assertEqual('f1-result', r)
                self.assertEqual('', file.getvalue())
            finally:

                file = None
        finally:

            enable_tracing(tracing_enabled)


    def test__trace__WITH_TRACING_ENABLED(self):

        tracing_enabled = True if enable_tracing(True) else False

        try:

            set_program_name('myprog1')

            global file

            file = StringIO()

            try:

                r = f1()

                self.assertEqual('f1-result', r)
                self.assertRegex(file.getvalue(), r'^\[myprog1, \d+, \d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{6}, .*TRACE.*\]: f1()')
            finally:

                file = None
        finally:

            enable_tracing(tracing_enabled)


    def test__tracefunc__WITH_TRACING_ENABLED(self):

        with patch('sys.stderr', new=StringIO()) as fake_stderr:

            tracing_enabled = True if enable_tracing(True) else False

            try:

                set_program_name('myprog3')

                r = f2()

                self.assertEqual('f2-result', r)
                self.assertRegex(fake_stderr.getvalue(), r'^\[myprog3, \d+, \d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{6}, .*TRACE.*\]: f2()')
            finally:

                enable_tracing(tracing_enabled)




import setuptools

setuptools.setup(

    name='diagnosticism',
    version='0.15.0',

    author='Matt Wilson',
    author_email='matthew@synesis.com.au',
    classifiers=[

        'Intended Audience :: Developers',
        "License :: OSI Approved :: BSD License",
        'Natural Language :: English',
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
    ],
    description='Basic diagnostic facilities, for Python',
    keywords='Diagnostic Diagnostics Logging Trace Tracing Stopwatch',
    license='BSD-3-Clause',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    packages=[
        'diagnosticism',
        'diagnosticism.internal',
        'examples',
        'tests',
    ],
    url='https://github.com/synesissoftware/diagnosticism.Python',
)

